/******************************************************************************
* Copyright (c) 2021, Martin Renou                                            *
*                                                                             *
* Distributed under the terms of the BSD 3-Clause License.                    *
*                                                                             *
* The full license is in the file LICENSE, distributed with this software.    *
*******************************************************************************/

#ifndef XCANVAS_CONFIG_CLING_HPP
#define XCANVAS_CONFIG_CLING_HPP

#pragma cling add_library_path("D:/bld/xcanvas_1649688611029/_h_env/Library/lib")
#pragma cling load("libxcanvas")

#endif
